#ifndef CONFIGDATAIMP_H
#define CONFIGDATAIMP_H

#include "ConfigData.h"

#include <memory>
#include <string>

class ConfigDataImp : public ConfigData
{
public:
    static std::shared_ptr<ConfigDataImp> getNewInstance();

    real getViscosity() override;
    uint getNumberOfDevices() override;
    std::vector<uint> getDevices() override;
    std::string getOutputPath() override;
    std::string getPrefix() override;
    std::string getGridPath() override;
    bool getPrintOutputFiles() override;
    bool getGeometryValues() override;
    bool getCalc2ndOrderMoments() override;
    bool getCalc3rdOrderMoments() override;
    bool getCalcHighOrderMoments() override;
    bool getReadGeo() override;
    bool getCalcMedian() override;
    bool getCalcDragLift() override;
    bool getCalcCp() override;
    bool getWriteVeloASCIIfiles() override;
    bool getCalcPlaneConc() override;
    bool getConcFile() override;
    bool getStreetVelocityFile() override;
    bool getUseMeasurePoints() override;
    bool getUseWale() override;
    bool getUseInitNeq() override;
    bool getSimulatePorousMedia() override;
    uint getD3Qxx() override;
    uint getTEnd() override;
    uint getTOut() override;
    uint getTStartOut() override;
    uint getTimeCalcMedStart() override;
    uint getTimeCalcMedEnd() override;
    uint getPressInID() override;
    uint getPressOutID() override;
    uint getPressInZ() override;
    uint getPressOutZ() override;
    bool getDiffOn() override;
    uint getDiffMod() override;
    real getDiffusivity() override;
    real getTemperatureInit() override;
    real getTemperatureBC() override;
    real getVelocity() override;
    real getViscosityRatio() override;
    real getVelocityRatio() override;
    real getDensityRatio() override;
    real getPressRatio() override;
    real getRealX() override;
    real getRealY() override;
    real getFactorPressBC() override;
    std::string getGeometryFileC() override;
    std::string getGeometryFileM() override;
    std::string getGeometryFileF() override;
    uint getClockCycleForMP() override;
    uint getTimestepForMP() override;
    real getForcingX() override;
    real getForcingY() override;
    real getForcingZ() override;
    real getQuadricLimiterP() override;
    real getQuadricLimiterM() override;
    real getQuadricLimiterD() override;
    bool getCalcParticles() override;
    int getParticleBasicLevel() override;
    int getParticleInitLevel() override;
    int getNumberOfParticles() override;
    real getStartXHotWall() override;
    real getEndXHotWall() override;
    std::vector<std::string> getPossNeighborFilesX() override;
    std::vector<std::string> getPossNeighborFilesY() override;
    std::vector<std::string> getPossNeighborFilesZ() override;
    // std::vector<std::string> getPossNeighborFilesX();
    // std::vector<std::string> getPossNeighborFilesY();
    // std::vector<std::string> getPossNeighborFilesZ();
    int getTimeDoCheckPoint() override;
    int getTimeDoRestart() override;
    bool getDoCheckPoint() override;
    bool getDoRestart() override;
    uint getMaxLevel() override;
    std::vector<int> getGridX() override;
    std::vector<int> getGridY() override;
    std::vector<int> getGridZ() override;
    std::vector<int> getDistX() override;
    std::vector<int> getDistY() override;
    std::vector<int> getDistZ() override;
    std::vector<bool> getNeedInterface() override;
    std::string getMainKernel() override;
    bool getMultiKernelOn() override;
    std::vector<int> getMultiKernelLevel() override;
    std::vector<std::string> getMultiKernelName() override;

    void setViscosity(real viscosity);
    void setNumberOfDevices(uint numberOfDevices);
    void setDevices(std::vector<uint> devices);
    void setOutputPath(std::string outputPath);
    void setPrefix(std::string prefix);
    void setGridPath(std::string gridPath);
    void setPrintOutputFiles(bool printOutputFiles);
    void setGeometryValues(bool geometryValues);
    void setCalc2ndOrderMoments(bool calc2ndOrderMoments);
    void setCalc3rdOrderMoments(bool calc3rdOrderMoments);
    void setCalcHighOrderMoments(bool calcHighOrderMoment);
    void setReadGeo(bool readGeo);
    void setCalcMedian(bool calcMedian);
    void setCalcDragLift(bool calcDragLift);
    void setCalcCp(bool calcCp);
    void setWriteVeloASCIIfiles(bool writeVeloASCIIfiles);
    void setCalcPlaneConc(bool calcPlaneConc);
    void setConcFile(bool concFile);
    void setStreetVelocityFile(bool streetVelocityFile);
    void setUseMeasurePoints(bool useMeasurePoints);
    void setUseWale(bool useWale);
    void setUseInitNeq(bool useInitNeq);
    void setSimulatePorousMedia(bool simulatePorousMedia);
    void setD3Qxx(uint d3Qxx);
    void setTEnd(uint tEnd);
    void setTOut(uint tOut);
    void setTStartOut(uint tStartOut);
    void setTimeCalcMedStart(uint timeCalcMedStart);
    void setTimeCalcMedEnd(uint timeCalcMedEnd);
    void setPressInID(uint pressInID);
    void setPressOutID(uint pressOutID);
    void setPressInZ(uint pressInZ);
    void setPressOutZ(uint pressOutZ);
    void setDiffOn(bool diffOn);
    void setDiffMod(uint diffMod);
    void setDiffusivity(real diffusivity);
    void setTemperatureInit(real temperatureInit);
    void setTemperatureBC(real temperatureBC);
    // void setViscosity(real viscosity);
    void setVelocity(real velocity);
    void setViscosityRatio(real viscosityRatio);
    void setVelocityRatio(real velocityRatio);
    void setDensityRatio(real fensityRatio);
    void setPressRatio(real pressRatio);
    void setRealX(real realX);
    void setRealY(real realY);
    void setFactorPressBC(real factorPressBC);
    void setGeometryFileC(std::string geometryFileC);
    void setGeometryFileM(std::string geometryFileM);
    void setGeometryFileF(std::string geometryFileF);
    void setClockCycleForMP(uint clockCycleForMP);
    void setTimestepForMP(uint timestepForMP);
    void setForcingX(real forcingX);
    void setForcingY(real forcingY);
    void setForcingZ(real forcingZ);
    void setQuadricLimiterP(real quadricLimiterP);
    void setQuadricLimiterM(real quadricLimiterM);
    void setQuadricLimiterD(real quadricLimiterD);
    void setCalcParticles(bool calcParticles);
    void setParticleBasicLevel(int particleBasicLevel);
    void setParticleInitLevel(int particleInitLevel);
    void setNumberOfParticles(int numberOfParticles);
    void setStartXHotWall(real startXHotWall);
    void setEndXHotWall(real endXHotWall);
    void setPossNeighborFilesX(std::vector<std::string> possNeighborFilesX);
    void setPossNeighborFilesY(std::vector<std::string> possNeighborFilesY);
    void setPossNeighborFilesZ(std::vector<std::string> possNeighborFilesZ);
    // void setPossNeighborFilesX(std::vector<std::string> possNeighborFilesX);
    // void setPossNeighborFilesY(std::vector<std::string> possNeighborFilesY);
    // void setPossNeighborFilesZ(std::vector<std::string> possNeighborFilesZ);
    void setTimeDoCheckPoint(int timeDoCheckPoint);
    void setTimeDoRestart(int timeDoRestart);
    void setDoCheckPoint(bool doCheckPoint);
    void setDoRestart(bool doRestart);
    void setMaxLevel(uint maxLevel);
    void setGridX(std::vector<int> gridX);
    void setGridY(std::vector<int> gridY);
    void setGridZ(std::vector<int> gridZ);
    void setDistX(std::vector<int> distX);
    void setDistY(std::vector<int> distY);
    void setDistZ(std::vector<int> distZ);
    void setNeedInterface(std::vector<bool> needInterface);
    void setMainKernel(std::string mainKernel);
    void setMultiKernelOn(bool multiKernelOn);
    void setMultiKernelLevel(std::vector<int> multiKernelLevel);
    void setMultiKernelName(std::vector<std::string> multiKernelName);

    bool isViscosityInConfigFile() override;
    bool isNumberOfDevicesInConfigFile() override;
    bool isDevicesInConfigFile() override;
    bool isOutputPathInConfigFile() override;
    bool isPrefixInConfigFile() override;
    bool isGridPathInConfigFile() override;
    bool isPrintOutputFilesInConfigFile() override;
    bool isGeometryValuesInConfigFile() override;
    bool isCalc2ndOrderMomentsInConfigFile() override;
    bool isCalc3rdOrderMomentsInConfigFile() override;
    bool isCalcHighOrderMomentsInConfigFile() override;
    bool isReadGeoInConfigFile() override;
    bool isCalcMedianInConfigFile() override;
    bool isCalcDragLiftInConfigFile() override;
    bool isCalcCpInConfigFile() override;
    bool isWriteVeloASCIIfilesInConfigFile() override;
    bool isCalcPlaneConcInConfigFile() override;
    bool isConcFileInConfigFile() override;
    bool isStreetVelocityFileInConfigFile() override;
    bool isUseMeasurePointsInConfigFile() override;
    bool isUseWaleInConfigFile() override;
    bool isUseInitNeqInConfigFile() override;
    bool isSimulatePorousMediaInConfigFile() override;
    bool isD3QxxInConfigFile() override;
    bool isTEndInConfigFile() override;
    bool isTOutInConfigFile() override;
    bool isTStartOutInConfigFile() override;
    bool isTimeCalcMedStartInConfigFile() override;
    bool isTimeCalcMedEndInConfigFile() override;
    bool isPressInIDInConfigFile() override;
    bool isPressOutIDInConfigFile() override;
    bool isPressInZInConfigFile() override;
    bool isPressOutZInConfigFile() override;
    bool isDiffOnInConfigFile() override;
    bool isDiffModInConfigFile() override;
    bool isDiffusivityInConfigFile() override;
    bool isTemperatureInitInConfigFile() override;
    bool isTemperatureBCInConfigFile() override;
    // bool isViscosityInConfigFile();
    bool isVelocityInConfigFile() override;
    bool isViscosityRatioInConfigFile() override;
    bool isVelocityRatioInConfigFile() override;
    bool isDensityRatioInConfigFile() override;
    bool isPressRatioInConfigFile() override;
    bool isRealXInConfigFile() override;
    bool isRealYInConfigFile() override;
    bool isFactorPressBCInConfigFile() override;
    bool isGeometryFileCInConfigFile() override;
    bool isGeometryFileMInConfigFile() override;
    bool isGeometryFileFInConfigFile() override;
    bool isClockCycleForMPInConfigFile() override;
    bool isTimestepForMPInConfigFile() override;
    bool isForcingXInConfigFile() override;
    bool isForcingYInConfigFile() override;
    bool isForcingZInConfigFile() override;
    bool isQuadricLimiterPInConfigFile() override;
    bool isQuadricLimiterMInConfigFile() override;
    bool isQuadricLimiterDInConfigFile() override;
    bool isCalcParticlesInConfigFile() override;
    bool isParticleBasicLevelInConfigFile() override;
    bool isParticleInitLevelInConfigFile() override;
    bool isNumberOfParticlesInConfigFile() override;
    bool isNeighborWSBInConfigFile() override;
    bool isStartXHotWallInConfigFile() override;
    bool isEndXHotWallInConfigFile() override;
    bool isPossNeighborFilesXInConfigFile() override;
    bool isPossNeighborFilesYInConfigFile() override;
    bool isPossNeighborFilesZInConfigFile() override;
    bool isTimeDoCheckPointInConfigFile() override;
    bool isTimeDoRestartInConfigFile() override;
    bool isDoCheckPointInConfigFile() override;
    bool isDoRestartInConfigFile() override;
    bool isMaxLevelInConfigFile() override;
    bool isGridXInConfigFile() override;
    bool isGridYInConfigFile() override;
    bool isGridZInConfigFile() override;
    bool isDistXInConfigFile() override;
    bool isDistYInConfigFile() override;
    bool isDistZInConfigFile() override;
    bool isNeedInterfaceInConfigFile() override;
    bool isMainKernelInConfigFile() override;
    bool isMultiKernelOnInConfigFile() override;
    bool isMultiKernelLevelInConfigFile() override;
    bool isMultiKernelNameInConfigFile() override;

private:
    ConfigDataImp();

    real viscosity;
    uint numberOfDevices;
    std::vector<uint> devices;
    std::string outputPath;
    std::string prefix;
    std::string gridPath;
    bool printOutputFiles;
    bool geometryValues;
    bool calc2ndOrderMoments;
    bool calc3rdOrderMoments;
    bool calcHighOrderMoments;
    bool readGeo;
    bool calcMedian;
    bool calcDragLift;
    bool calcCp;
    bool writeVeloASCIIfiles;
    bool calcPlaneConc;
    bool concFile;
    bool streetVelocityFile;
    bool useMeasurePoints;
    bool useWale;
    bool useInitNeq;
    bool simulatePorousMedia;
    uint d3Qxx;
    uint tEnd;
    uint tOut;
    uint tStartOut;
    uint timeCalcMedStart;
    uint timeCalcMedEnd;
    uint pressInID;
    uint pressOutID;
    uint pressInZ;
    uint pressOutZ;
    bool diffOn;
    uint diffMod;
    real diffusivity;
    real temperatureInit;
    real temperatureBC;
    // real viscosity;
    real velocity;
    real viscosityRatio;
    real velocityRatio;
    real densityRatio;
    real pressRatio;
    real realX;
    real realY;
    real factorPressBC;
    std::string geometryFileC;
    std::string geometryFileM;
    std::string geometryFileF;
    uint clockCycleForMP;
    uint timestepForMP;
    real forcingX;
    real forcingY;
    real forcingZ;
    real quadricLimiterP;
    real quadricLimiterM;
    real quadricLimiterD;
    bool calcParticles;
    int particleBasicLevel;
    int particleInitLevel;
    int numberOfParticles;
    real startXHotWall;
    real endXHotWall;
    std::vector<std::string> possNeighborFilesX;
    std::vector<std::string> possNeighborFilesY;
    std::vector<std::string> possNeighborFilesZ;
    // std::vector<std::string> possNeighborFilesX;
    // std::vector<std::string> possNeighborFilesY;
    // std::vector<std::string> possNeighborFilesZ;
    int timeDoCheckPoint;
    int timeDoRestart;
    bool doCheckPoint;
    bool doRestart;
    int maxLevel;
    std::vector<int> gridX;
    std::vector<int> gridY;
    std::vector<int> gridZ;
    std::vector<int> distX;
    std::vector<int> distY;
    std::vector<int> distZ;
    std::vector<bool> needInterface;
    std::string mainKernel;
    bool multiKernelOn;
    std::vector<int> multiKernelLevel;
    std::vector<std::string> multiKernelName;

    bool isViscosity;
    bool isNumberOfDevices;
    bool isDevices;
    bool isOutputPath;
    bool isPrefix;
    bool isGridPath;
    bool isPrintOutputFiles;
    bool isGeometryValues;
    bool isCalc2ndOrderMoments;
    bool isCalc3rdOrderMoments;
    bool isCalcHighOrderMoments;
    bool isReadGeo;
    bool isCalcMedian;
    bool isCalcDragLift;
    bool isCalcCp;
    bool isWriteVeloASCII;
    bool isCalcPlaneConc;
    bool isConcFile;
    bool isStreetVelocityFile;
    bool isUseMeasurePoints;
    bool isUseWale;
    bool isUseInitNeq;
    bool isSimulatePorousMedia;
    bool isD3Qxx;
    bool isTEnd;
    bool isTOut;
    bool isTStartOut;
    bool isTimeCalcMedStart;
    bool isTimeCalcMedEnd;
    bool isPressInID;
    bool isPressOutID;
    bool isPressInZ;
    bool isPressOutZ;
    bool isDiffOn;
    bool isDiffMod;
    bool isDiffusivity;
    bool isTemperatureInit;
    bool isTemperatureBC;
    // bool isViscosity;
    bool isVelocity;
    bool isViscosityRatio;
    bool isVelocityRatio;
    bool isDensityRatio;
    bool isPressRatio;
    bool isRealX;
    bool isRealY;
    bool isFactorPressBC;
    bool isGeometryFileC;
    bool isGeometryFileM;
    bool isGeometryFileF;
    bool isClockCycleForMP;
    bool isTimestepForMP;
    bool isForcingX;
    bool isForcingY;
    bool isForcingZ;
    bool isQuadricLimiterP;
    bool isQuadricLimiterM;
    bool isQuadricLimiterD;
    bool isCalcParticles;
    bool isParticleBasicLevel;
    bool isParticleInitLevel;
    bool isNumberOfParticles;
    bool isNeighborWSB;
    bool isStartXHotWall;
    bool isEndXHotWall;
    bool isPossNeighborFilesX;
    bool isPossNeighborFilesY;
    bool isPossNeighborFilesZ;
    bool isTimeDoCheckPoint;
    bool isTimeDoRestart;
    bool isDoCheckPoint;
    bool isDoRestart;
    bool isMaxLevel;
    bool isGridX;
    bool isGridY;
    bool isGridZ;
    bool isDistX;
    bool isDistY;
    bool isDistZ;
    bool isNeedInterface;
    bool isMainKernel;
    bool isMultiKernelOn;
    bool isMultiKernelLevel;
    bool isMultiKernelName;
};
#endif
